/*
 * manhattan - pu-lib.c
 *
 * Copyright (C) 2012  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * major part taken from git usage.c and wrapper.c
 * Copyright (C) Linus Torvalds, 2005
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <pu-lib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>

static int dying;

static NORETURN void die_builtin(const char *err, va_list params);

static NORETURN_PTR void (*die_routine)(const char *err, va_list params) = die_builtin;

static void vreportf(const char *prefix, const char *err, va_list params)
{
	char msg[4096];
	vsnprintf(msg, sizeof(msg), err, params);
	fprintf(stderr, "%s%s\n", prefix, msg);
}

static NORETURN void die_builtin(const char *err, va_list params)
{
	vreportf("fatal: ", err, params);
	exit(128);
}

void *xmalloc(size_t size)
{
	void *p = malloc(size);

	if (unlikely(!p))
		die("cannot allocate %d bytes of memory\n", size);
	return p;
}

void xfree(const void *ptr)
{
	/*
	 * we use cast to integer to avoid bogus gcc warning, caused
	 * by brain-damaged free() in C99.
	 */
	free((void *)(uintptr_t)ptr);
}

/*
 * xread() is the same a read(), but it automatically restarts read()
 * operations with a recoverable error (EINTR). xread()
 * DOES NOT GUARANTEE that "len" bytes is read even if the data is available.
 */
ssize_t xread(int fd, void *buf, size_t len)
{
	ssize_t nr;
	while (1) {
		nr = read(fd, buf, len);
		if ((nr < 0) && (errno == EINTR))
			continue;
		return nr;
	}
}

/*
 * xwrite() is the same a write(), but it automatically restarts write()
 * operations with a recoverable error (EINTR). xwrite() DOES NOT
 * GUARANTEE that "len" bytes is written even if the operation is successful.
 */
ssize_t xwrite(int fd, const void *buf, size_t len)
{
	ssize_t nr;
	while (1) {
		nr = write(fd, buf, len);
		if ((nr < 0) && (errno == EINTR))
			continue;
		return nr;
	}
}

ssize_t read_in_full(int fd, void *buf, size_t count)
{
	char *p = buf;
	ssize_t total = 0;

	while (count > 0) {
		ssize_t loaded = xread(fd, p, count);
		if (loaded < 0)
			return -1;
		if (loaded == 0)
			return total;
		count -= loaded;
		p += loaded;
		total += loaded;
	}

	return total;
}

ssize_t write_in_full(int fd, const void *buf, size_t count)
{
	const char *p = buf;
	ssize_t total = 0;

	while (count > 0) {
		ssize_t written = xwrite(fd, p, count);
		if (written < 0)
			return -1;
		if (!written) {
			errno = ENOSPC;
			return -1;
		}
		count -= written;
		p += written;
		total += written;
	}

	return total;
}

void NORETURN die(const char *err, ...)
{
	va_list params;

	if (dying) {
		fputs("fatal: recursion detected in die handler\n", stderr);
		exit(128);
	}
	dying = 1;

	va_start(params, err);
	die_routine(err, params);
	va_end(params);
}

void NORETURN die_errno(const char *fmt, ...)
{
	va_list params;
	char fmt_with_err[1024];
	char str_error[256], *err;
	int i, j;

	if (dying) {
		fputs("fatal: recursion detected in die_errno handler\n",
			stderr);
		exit(128);
	}
	dying = 1;

	err = strerror(errno);
	for (i = j = 0; err[i] && j < sizeof(str_error) - 1; ) {
		if ((str_error[j++] = err[i++]) != '%')
			continue;
		if (j < sizeof(str_error) - 1) {
			str_error[j++] = '%';
		} else {
			/* No room to double the '%', so we overwrite it with
			 * '\0' below */
			j--;
			break;
		}
	}
	str_error[j] = 0;
	snprintf(fmt_with_err, sizeof(fmt_with_err), "%s: %s", fmt, str_error);

	va_start(params, fmt);
	die_routine(fmt_with_err, params);
	va_end(params);
}

void warn(const char *fmt, ...)
{
	va_list params;

	va_start(params, fmt);
	vreportf("warning: ", fmt, params);
	va_end(params);
}

void info(const char *fmt, ...)
{
	va_list params;

	va_start(params, fmt);
	vreportf("", fmt, params);
	va_end(params);
}

void write_or_die(int fd, const void *buf, size_t count)
{
	if (write_in_full(fd, buf, count) < 0) {
		if (errno == EPIPE)
			exit(0);
		die_errno("write error");
	}
}
