/*
 * manhattan - rtunit_comm.c
 *
 * Communication with DSP
 *
 * Copyright (C) 2012  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <sys/mman.h>
#include <asm_io.h>
#include <crc14.h>
#include <gcp.h>
#include <pu-lib.h>
#include <pu-proto.h>
#include <pu-proto-lib.h>
#include <pugcp-proto.h>
#include <pugcp-proto-lib.h>
#include <pu-buff.h>
#include <mclock.h>
#include <limits.h>
#include <asm_types.h>
#include <termios.h>

#define GRTP_NODES		16
#define GRTP_NODE_MASK		(GRTP_NODES - 1)
#define GCP_SLOTS		4
#define GRTP_FRAME_LEN		10

enum gcp_state {
	GCP_CLOSED		= 0,
	GCP_RST_SENDING,
	GCP_RST_SENT,
	GCP_ESTABLISHED
};

static struct list_head request_queue[GRTP_NODES];
static struct pu_buff *request_current[GRTP_NODES][GCP_SLOTS];
static u8 request_frame[GRTP_NODES][GCP_SLOTS][GRTP_FRAME_LEN];
static unsigned long long request_xmit_time[GRTP_NODES][GCP_SLOTS];
static unsigned int request_seq[GRTP_NODES];
static unsigned int request_xmit[GRTP_NODES];
static unsigned int request_ack[GRTP_NODES];
static enum gcp_state gcp_state[GRTP_NODES];
static u8 gcp_rst_frame[GRTP_NODES][GRTP_FRAME_LEN];
static u32 gcp_rst_magic[GRTP_NODES];
static u32 gcp_timeouts[GRTP_NODES];

static unsigned int request_node;
static unsigned int request_bytes;

#define le32_to_cpu(x)	(x)
#define cpu_to_le32(x)	(x)

#define GRTP_PAYLOAD_LEN	6
#define GRTP_SYNC		0x80

#define GRTP_CHECKSUM_INIT	0x3fff
#define GRTP_CHECKSUM_CHECK	0x3c90

#define GRTP_BUF_SIZE	8

struct grtp_frame {
	u32 frame[(GRTP_FRAME_LEN + 3) / 4];
};

struct grtp {
	unsigned int len;
	u32 frame[(GRTP_FRAME_LEN + 3) / 4];
	u16 crc;

	unsigned char rx_buf[GRTP_BUF_SIZE];
	unsigned int rx_head;
	unsigned int rx_tail;
};

static struct grtp grtp;

static inline u32 grtp_decode_word(u32 x, unsigned int bits)
{
	static const u32 decode_tab[16] = {
		0x00000000,	/* 0x0 */
		0x00000080,	/* 0x1 */
		0x00008000,	/* 0x2 */
		0x00008080,	/* 0x3 */
		0x00800000,	/* 0x4 */
		0x00800080,	/* 0x5 */
		0x00808000,	/* 0x6 */
		0x00808080,	/* 0x7 */
		0x80000000,	/* 0x8 */
		0x80000080,	/* 0x9 */
		0x80008000,	/* 0xa */
		0x80008080,	/* 0xb */
		0x80800000,	/* 0xc */
		0x80800080,	/* 0xd */
		0x80808000,	/* 0xe */
		0x80808080,	/* 0xf */
	};

	return x + decode_tab[bits];
}

struct pu_buff *gcp_psched_ack(unsigned int command, unsigned int payload)
{
	struct pugcp_psched *p;
	struct pu_buff *skb;
	unsigned int id;
	unsigned int is_write;
	unsigned int value;

	skb = pugcp_proto_alloc_skb(sizeof(*p));
	p = (void *) pu_skb_put(skb, sizeof(*p));

	is_write = (payload & GCP_PSCHED_WRITE) != 0;
	id = (payload >> GCP_PSCHED_ID_SHIFT) & GCP_PSCHED_ID_MASK;
	value = (payload >> GCP_PSCHED_VALUE_SHIFT) & GCP_PSCHED_VALUE_MASK;

	if (is_write)
		id |= PUGCP_PSCHED_ID_WRITE;

	p->id = cpu_to_be32(id);
	p->value = cpu_to_be32(value);
	return skb;	
}

static void gcp_receive_connect_ack(unsigned int hdr, const u32 *frame)
{
	unsigned int node;
	unsigned int magic;
	unsigned int seq;

	node = ((hdr >> GCP_NODE_SHIFT) & GCP_NODE_MASK);

	if (!(hdr & GCP_SEQCMD_RST)) {
		warn("gcp: unexpected ack without rst");
		return;
	}

	seq = (hdr >> GCP_SEQCMD_SEQ_SHIFT) & GCP_SEQCMD_SEQ_MASK;
	if (seq != ((request_ack[node] - 1) & GCP_SEQCMD_SEQ_MASK)) {
		warn("gcp: unexpected ack+rst sequence number");
		return;
	}

	magic = le32_to_cpu(frame[GCP_PAYLOAD]);
	if (magic != gcp_rst_magic[node]) {
		warn("gcp: unexpected ack+rst magic: %08x should be %08x",
				magic, gcp_rst_magic[node]);
		if (magic != 0xffffda03)
			return;
		warn("old rtunit firmware without echo found, continuing");
	}

	info("established GCP connection with node %d", node);
	gcp_state[node] = GCP_ESTABLISHED;
}

static void gcp_receive_seqcmd_ack(unsigned int hdr, const u32 *frame)
{
	struct pu_buff *skb, *req;
	struct pugcp_header *g;
	unsigned int node;
	unsigned int seq;
	unsigned int command;
	u32 payload;
	void *priv;
	void **cb;

	node = ((hdr >> GCP_NODE_SHIFT) & GCP_NODE_MASK);
	info("got ACK: %06x", hdr);
	switch (gcp_state[node]) {
	case GCP_CLOSED:
	case GCP_RST_SENDING:
		warn("gcp: ACK on closed connection");
		return;
	case GCP_RST_SENT:
		gcp_receive_connect_ack(hdr, frame);
		return;
	case GCP_ESTABLISHED:
		break;
	}

	if (hdr & GCP_SEQCMD_RST) {
		warn("gcp: unexpected ack+rst, already connected");
		return;
	}

	seq = (hdr >> GCP_SEQCMD_SEQ_SHIFT) & GCP_SEQCMD_SEQ_MASK;
	if (request_ack[node] == request_xmit[node]) {
		warn("gcp: unexpected ack, already acknowledged");
		return;
	}
	if ((request_ack[node] & GCP_SEQCMD_SEQ_MASK) != seq) {
		warn("gcp: invalid ack, lost frame");
		return;
	}
	request_ack[node]++;
	gcp_timeouts[node] = 0;
	req = request_current[node][seq];
	if (!req) {
		warn("gcp: no skb");
		return;
	}
	request_current[node][seq] = NULL;

	cb = (void **) req->cb;
	priv = *cb;

	command = (hdr >> GCP_SEQCMD_CMD_SHIFT) & GCP_SEQCMD_CMD_MASK;

	payload = le32_to_cpu(frame[GCP_PAYLOAD]);
	info("got response: payload: %08x", payload);
	if (command >= GCP_CPSCHED_SYNC
			&& command <= GCP_CPSCHED_START_AFTER) {
		skb = gcp_psched_ack(command, payload);
	} else {
		pu_proto_error(priv, req, PU_EINVAL);
		return;
	}
	g = (void *) pu_skb_push(skb, sizeof(*g));
	memset(g, 0, sizeof(*g));
	g->checksum = be32_to_cpu(0);
	g->proto = be32_to_cpu(0);
	g->node = be32_to_cpu(node);
	g->command = be32_to_cpu(command);

	pu_dump_skb(skb);
	pu_proto_xmit_resp(priv, skb, req);
}

extern void pevent_callback(unsigned int pos, unsigned int events);

static void gcp_pevent_info(unsigned int cmd, u32 arg)
{
	unsigned int position;
	unsigned int outputs;

	position = (arg >> GCP_EVENT_VALUE_SHIFT) & GCP_EVENT_VALUE_MASK;
	position <<= 32 - GCP_EVENT_VALUE_BITS;
	outputs = (arg >> GCP_EVENT_OUTPUT_SHIFT) & GCP_EVENT_OUTPUT_MASK;

	pevent_callback(position, outputs);
}

static void gcp_receive_seqcmd_cmd(unsigned int hdr, const u32 *frame)
{
	unsigned int node;
	unsigned int cmd;
	unsigned int arg;

	node = ((hdr >> GCP_NODE_SHIFT) & GCP_NODE_MASK);

	cmd = (hdr >> GCP_SEQCMD_CMD_SHIFT) & GCP_SEQCMD_CMD_MASK;
	arg = le32_to_cpu(frame[GCP_PAYLOAD]);

	if (node == 15) {
		if (cmd == GCP_CPEVENT_INFO)
			gcp_pevent_info(cmd, arg);
	}
}

static void gcp_receive_frame(const u32 *frame)
{
	unsigned int hdr;
	unsigned int proto;

	hdr = le32_to_cpu(frame[GCP_HEADER]);
	proto = ((hdr >> GCP_PROTO_SHIFT) & GCP_PROTO_MASK);
	if (proto == GCP_PROTO_SEQCMD) {
		if (hdr & GCP_SEQCMD_ACK)
			gcp_receive_seqcmd_ack(hdr, frame);
		else
			gcp_receive_seqcmd_cmd(hdr, frame);
	}
}

static void grtp_receive_frame(struct grtp *g)
{
	u32 *frame = g->frame;
	unsigned int a;
	unsigned int b;
	unsigned int bits;
	unsigned int head;
	unsigned int next;

	a = le32_to_cpu(frame[0]);
	b = le32_to_cpu(frame[1]);
	a &= ~GRTP_SYNC;
	bits = b >> 24;
	a = grtp_decode_word(a, bits & 0xf);
	b = grtp_decode_word(b, bits >> 4);
	frame[0] = cpu_to_le32(a);
	frame[1] = cpu_to_le32(b);

	head = g->rx_head;
	next = (head + 1) & (GRTP_BUF_SIZE - 1);
	if (next != g->rx_tail) {
		g->rx_buf[head] = ((u8 *) frame)[6];
		g->rx_head = next;
	}

	gcp_receive_frame(frame);

//	g->callback(&g->tty, frame);
}

static void grtp_receive_buf(struct grtp *g, const unsigned char *cp,
		int count)
{
	unsigned int len = g->len;

	if (!count)
		return;

	do {
		unsigned char c;
		u8 *frame = (u8 *) g->frame;

		c = *cp;
		if (c & GRTP_SYNC) {
			len = 0;
			frame[len++] = c;
			g->crc = crc14_byte(GRTP_CHECKSUM_INIT, c);
		} else if (len) {
			frame[len++] = c;
			g->crc = crc14_byte(g->crc, c);
			if (len >= GRTP_FRAME_LEN) {
				if (g->crc == GRTP_CHECKSUM_CHECK)
					grtp_receive_frame(g);
				len = 0;
			}
		}
		cp++;
		count--;
	} while (count > 0);
	g->len = len;
}

static void grtp_encode(u8 *msg, const u8 *frame)
{
	unsigned int bits = 0;
	unsigned int crc;
	unsigned int i;

	i = 6;
	do {
		bits <<= 1;
		msg[i] = frame[i] & 0x7f;
		if (frame[i] & 0x80)
			bits |= 0x01;
	} while (i-- > 0);
	msg[0] |= GRTP_SYNC;
	msg[7] = bits;

	crc = crc14(GRTP_CHECKSUM_INIT, msg, GRTP_FRAME_LEN - 2);
	crc = ~crc;
	msg[GRTP_FRAME_LEN - 2] = (crc >> 0) & 0x7f;
	msg[GRTP_FRAME_LEN - 1] = (crc >> 7) & 0x7f;
}

static u32 grader_checksum(const u32 *data, unsigned int count)
{
	u32 sum = 0;
	unsigned int i;

	for (i = 0; i < count >> 2; i++)
		sum += be32_to_cpu(*data++);
	return sum;
}

static int gcp_format_payload_param(unsigned int command, struct pu_buff *skb,
		u32 *payload)
{
	struct pugcp_psched *p;
	unsigned int id;
	unsigned int value;
	int is_write; 

	if (skb->len < sizeof(*p))
		return -EINVAL;

	p = (void *) skb->data;
	id = be32_to_cpu(p->id);
	value = be32_to_cpu(p->value);
	fprintf(stderr, "x: %d %d\n", id, value);

	is_write = (id & PUGCP_PSCHED_ID_WRITE) != 0;
	id &= PUGCP_PSCHED_ID_MASK;
	if (id > GCP_PSCHED_ID_MASK)
		return -EINVAL;
	if (value > GCP_PSCHED_VALUE_MASK)
		return -EINVAL;
	*payload = id << GCP_PSCHED_ID_SHIFT;
	*payload |= value << GCP_PSCHED_VALUE_SHIFT;
	if (is_write)
		*payload |= GCP_PSCHED_WRITE;
	fprintf(stderr, "payload: %08x\n", *payload);
	return 0;
}

static int gcp_format_payload_cadd_event(unsigned int command,
		struct pu_buff *skb, u32 *payload)
{
	struct pugcp_cadd_event *p;
	unsigned int output;
	unsigned int position;

	if (skb->len < sizeof(*p))
		return -EINVAL;

	p = (void *) skb->data;
	output = be32_to_cpu(p->output);
	position = be32_to_cpu(p->position);
	fprintf(stderr, "queue '%d' at position %d\n", output, position);

	if (output > 255)
		return -EINVAL;

	position += (1 << (32 - GCP_PSCHED_VALUE_BITS)) / 2;
	position >>= 32 - GCP_PSCHED_VALUE_BITS;

	*payload = output << GCP_EVENT_OUTPUT_SHIFT;
	*payload |= position << GCP_EVENT_VALUE_SHIFT;
	fprintf(stderr, "payload: %08x\n", *payload);
	return 0;
}

static int gcp_format_payload(unsigned int command, struct pu_buff *skb,
		u32 *payload)
{
	if (command >= GCP_CPSCHED_SYNC
			&& command <= GCP_CPSCHED_START_AFTER) {
		return gcp_format_payload_param(command, skb, payload);
	} else if (command == GCP_CADD_EVENT) {
		return gcp_format_payload_cadd_event(command, skb, payload);
	} else {
		return -EINVAL;
	}
}

static int gcp_format_msg(u32 *frame, unsigned int seq,
		struct pu_buff *skb)
{
	struct pugcp_header *g;
	unsigned int command;
	unsigned int proto;
	unsigned int node;
	u32 payload = 0;
	int ret;
	u32 hdr;


	g = (void *) skb->data;
	pu_skb_pull(skb, sizeof(*g));

	/*
	 * values have been already checked
	 */
	proto = be32_to_cpu(g->proto);
	node = be32_to_cpu(g->node);
	command = be32_to_cpu(g->command);

	fprintf(stderr, "format2:\n");
	pu_dump_skb(skb);
	/*
	 * fast path, for some late errors
	 */
	ret = gcp_format_payload(command, skb, &payload);
	if (ret)
		goto out;
	fprintf(stderr, "after:\n");
	pu_dump_skb(skb);

	hdr = proto << GCP_PROTO_SHIFT;
	hdr |= seq << GCP_SEQCMD_SEQ_SHIFT;
	hdr |= node << GCP_NODE_SHIFT;
	hdr |= command << GCP_SEQCMD_CMD_SHIFT;
	frame[GCP_HEADER] = cpu_to_le32(hdr);
	frame[GCP_PAYLOAD] = cpu_to_le32(payload);
out:
	pu_skb_push(skb, sizeof(*g));
	return ret;
}
	
static void rtunit_comm_schedule(unsigned int node)
{
	if (list_empty(&request_queue[node]))
		return;

	while ((int) request_seq[node] - (int) request_ack[node] < 4) {
		u32 gcp_frame[2];
		unsigned int slot = request_seq[node] & (GCP_SLOTS - 1);
		struct pu_buff *skb;
		int ret;

		if (request_current[node][slot])
			break;

		skb = list_entry(request_queue[node].next, struct pu_buff, list);
		list_del(&skb->list);

		fprintf(stderr, "formating:\n");
		pu_dump_skb(skb);
		ret = gcp_format_msg(gcp_frame, slot, skb);
		fprintf(stderr, "after format:\n");
		pu_dump_skb(skb);
		if (ret < 0) {
			void *priv;
			void **cb;
			cb = (void **) skb->cb;
			priv = *cb;

			pu_proto_error_errno(priv, skb, -ret);
			if (list_empty(&request_queue[node]))
				return;
			continue;
		}


		request_current[node][slot] = skb;

		fprintf(stderr, "new frame: %08x %08x\n",
				gcp_frame[0], gcp_frame[1]);
		grtp_encode(request_frame[node][slot], (void *) gcp_frame);
		fprintf(stderr, "raw frame: %02x %02x %02x %02x %02x %02x %02x  %02x  %02x %02x\n",
				request_frame[node][slot][0],
				request_frame[node][slot][1],
				request_frame[node][slot][2],
				request_frame[node][slot][3],
				request_frame[node][slot][4],
				request_frame[node][slot][5],
				request_frame[node][slot][6],
				request_frame[node][slot][7],
				request_frame[node][slot][8],
				request_frame[node][slot][9]);
		request_seq[node]++;

		if (list_empty(&request_queue[node]))
			return;
	}
}

int rtunit_comm_timeout(void)
{
	unsigned long long t = time_ns();
	unsigned long long t_min = ULLONG_MAX;
	unsigned long delta;
	unsigned int node;

	for (node = 0; node < GRTP_NODES; node++) {
		enum gcp_state state;
		unsigned int slot;
		unsigned long timeout = 10000000;

		state = gcp_state[node];
		if (state == GCP_CLOSED || state == GCP_RST_SENDING)
			continue;

		if (state == GCP_ESTABLISHED) {
			if (request_xmit[node] == request_ack[node])
				continue;
			slot = request_ack[node] & (GCP_SLOTS - 1);
		} else {
			slot = 0;
		}

		if (node == 14)
			timeout = 2000000;
		if (t >= request_xmit_time[node][slot] + timeout) {
			if (state == GCP_ESTABLISHED) {
				fprintf(stderr, "timeout: %d -> %d\n",
						request_xmit[node],
						request_ack[node]);
				request_xmit[node] = request_ack[node];
				gcp_timeouts[node]++;
				if (gcp_timeouts[node] > 10) {
					warn("node %d: too many timeouts",
							node);
					gcp_state[node] = GCP_CLOSED;
				}
			} else {
				warn("node %d: RST timeout", node);
				gcp_state[node] = GCP_CLOSED;
			}
		} else if (t_min > request_xmit_time[node][slot] + timeout)
			t_min = request_xmit_time[node][slot] + timeout;
	}
	if (t_min == ULLONG_MAX)
		return -1;

	delta = (t_min - t + 1000000 - 1) / 1000000;
	return delta;
}

int rtunit_comm_want_out(void)
{
	unsigned int node;

	for (node = 0; node < GRTP_NODES; node++) {
		if (((int) request_xmit[node] - (int) request_seq[node]) < 0) {
			return (gcp_state[node] != GCP_RST_SENT);
		}
	}
	return 0;
}

static int rtunit_gcp_connect(int fd, unsigned int node)
{
	unsigned int bytes;
	unsigned int seq;
	unsigned int hdr;
	unsigned int command = GCP_CECHO;
	static u32 connect_magic;

	if (gcp_state[node] == GCP_CLOSED) {
		u32 gcp_frame[2];

		seq = (request_ack[node] - 1) & GCP_SEQCMD_SEQ_MASK;
		hdr = GCP_PROTO_SEQCMD << GCP_PROTO_SHIFT;
		hdr |= seq << GCP_SEQCMD_SEQ_SHIFT;
		hdr |= GCP_SEQCMD_RST;
		hdr |= node << GCP_NODE_SHIFT;
		hdr |= command << GCP_SEQCMD_CMD_SHIFT;
		gcp_frame[GCP_HEADER] = cpu_to_le32(hdr);
		gcp_frame[GCP_PAYLOAD] = cpu_to_le32(connect_magic);
		gcp_rst_magic[node] = connect_magic;
		grtp_encode(gcp_rst_frame[node], (void *) gcp_frame);
		connect_magic++;
		gcp_state[node] = GCP_RST_SENDING;
	}
	if (gcp_state[node] == GCP_RST_SENDING) {
		const u8 *frame = gcp_rst_frame[node];
		int ret;

		bytes = request_bytes;
		ret = write(fd, frame + bytes, GRTP_FRAME_LEN - bytes);
		if (ret < 0)
			return ret;
		bytes += ret;
		if (bytes >= GRTP_FRAME_LEN) {
			request_xmit_time[node][0] = time_ns();
			gcp_state[node] = GCP_RST_SENT;
			bytes = 0;
		}
		request_bytes = bytes;
	}
	return 0;
}

int rtunit_comm_pollout(int fd)
{
	const u8 *frame;
	unsigned int bytes;
	unsigned int node;
	unsigned int slot;
	int ret;

	node = request_node;
	while (((int) request_xmit[node] - (int) request_seq[node]) >= 0) {
		node = (node + 1) & GRTP_NODE_MASK;
		if (node == request_node)
			return 0;
	}
	request_node = node;

	if (gcp_state[node] != GCP_ESTABLISHED)
		return rtunit_gcp_connect(fd, node);

	slot = request_xmit[node] & (GCP_SLOTS - 1);
	frame = request_frame[node][slot];
	bytes = request_bytes;
	fprintf(stderr, "writting frame: %d bytes\n", GRTP_FRAME_LEN - bytes);
	ret = write(fd, frame + bytes, GRTP_FRAME_LEN - bytes);
	if (ret < 0)
		return ret;
	bytes += ret;
	if (bytes >= GRTP_FRAME_LEN) {
		request_xmit_time[node][slot] = time_ns();
		request_xmit[node]++;
		bytes = 0;
	}
	request_bytes = bytes;
	return 0;
}

int rtunit_comm_pollin(int fd)
{
	unsigned char buf[256];
	int ret;

	ret = read(fd, buf, sizeof(buf));
	if (!ret)
		exit(EXIT_SUCCESS);
	if (ret < 0)
		return -errno;

	grtp_receive_buf(&grtp, buf, ret);
	return 0;
}

int pu_proto_grader(void *priv, struct pu_buff *skb)
{
	struct pu_header *hdr = (void *) skb->data;
	struct pugcp_header *g;
	unsigned int data_bytes = be32_to_cpu(hdr->data_bytes);
	unsigned int command;
	unsigned int proto;
	unsigned int node;
	struct pu_buff *new;
	void **cb;

	if (data_bytes < 4)
		return -EINVAL;
	if (data_bytes & 3)
		return -EINVAL;
#if 0
	if (grader_checksum((void *) skb->data, data_bytes)) {
		fprintf(stderr, "invalid checksum\n");
		return -EINVAL;
	}
#endif

	pu_skb_pull(skb, sizeof(*hdr));
	if (data_bytes < sizeof(*g))
		return pu_proto_error(priv, skb, PU_EINVAL);

	g = (void *) skb->data;

	proto = be32_to_cpu(g->proto);
	if (proto >= 16)
		return pu_proto_error(priv, skb, PU_EINVAL);
	if (proto != GCP_PROTO_SEQCMD)
		return pu_proto_error(priv, skb, PU_EINVAL);

	node = be32_to_cpu(g->node);
	if (node >= 16)
		return pu_proto_error(priv, skb, PU_EINVAL);

	command = be32_to_cpu(g->command);
	if (command >= 4096)
		return pu_proto_error(priv, skb, PU_EINVAL);

	new = pu_skb_copy(skb);
	cb = (void **) new->cb;
	*cb = priv;
	fprintf(stderr, "scheduling:\n");
	pu_dump_skb(new);
	list_add_tail(&new->list, &request_queue[node]);
	rtunit_comm_schedule(node);
	return 0;
}

int rtunit_comm_init(const char *device)
{
	speed_t speed = B1500000;
	struct termios mode;
	unsigned int node;
	int fd;

	for (node = 0; node < GRTP_NODES; node++)
		INIT_LIST_HEAD(&request_queue[node]);

	fd = open(device, O_RDWR | O_NOCTTY);
	if (fd < 0) {
		fprintf(stderr, "can't open rtunit device '%s': %s\n",
				device, strerror(errno));
		return -errno;
	}

	tcgetattr(fd, &mode);
	cfmakeraw(&mode);
	mode.c_cflag |= CLOCAL | CREAD | CS8;
	mode.c_oflag = 0;
	cfsetispeed(&mode, speed);
	cfsetospeed(&mode, speed);
	tcflush(fd, TCIFLUSH);
	tcsetattr(fd, TCSANOW, &mode);

	if (fcntl(fd, F_SETFL, O_NONBLOCK))
		fprintf(stderr, "fcntl(): %s", strerror(errno));

	return fd;
}
